import { getMetadata, Serde } from '../decorator'
import { SerdeOptions, Stage } from '../options'
import { Constructor, staticImplements } from '../utils'
import { GenericVisitor } from './generic'
import { Deserialize, Deserializer } from './interface'

type DeserializeConstructor<T> = Deserialize<T> & { new(): Deserialize<T> }

function deserializeWith<T, D extends Deserializer, E extends DeserializeConstructor<T>>(deserializer: D, into: E, options: SerdeOptions): T {
  const visitor = new GenericVisitor<T>()
  const obj = deserializer.deserializeObject(visitor) as any
  const target = new into()
  const newObject = {} as any

  for (const property in target) {
    const name = options.getPropertyName(property, Stage.Deserialize)
    const value = obj[name] || options.getDefault(property)
    newObject[property] = value
    delete obj[name]
  }

  if (options.options.denyUnknownFields && Object.keys(obj).length > 0) {
    throw new TypeError(`Unexpected fields: ${Object.keys(obj).join(', ')}`)
  }

  return Object.assign(target, newObject)
}

export function deserialize<T, C extends Constructor>(constructor: C) {
  @staticImplements<Deserialize<T>>()
  class Deserializable extends constructor {
    static name = constructor.name

    static deserialize<D extends Deserializer>(deserializer: D): T {
      return deserializeWith(deserializer, this, getMetadata(this))
    }
  }

  // @ts-ignore
  Deserializable[Serde] = (constructor as any)[Serde]

  return Deserializable
}

